'use strict';

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

var objectEntries = function objectEntries(object) {
    return Object.keys(object).map(function (key) {
        return [key, object[key]];
    });
};
var matches = function matches(value) {
    return function (pattern) {
        return pattern.matches(value);
    };
};
var findAndBuild = function findAndBuild(pattern) {
    return findPattern(pattern).build(pattern);
};
var avaliblePatterns = [{
    build: function build(pattern) {
        return function (value) {
            return value === pattern;
        };
    },
    matches: function matches(pattern) {
        return typeof pattern === 'boolean' || typeof pattern === 'number' || typeof pattern === 'string' || pattern === null || pattern === undefined;
    }
}, {
    build: function build(pattern) {
        return function (value) {
            return value.getTime() === pattern.getTime();
        };
    },
    matches: function matches(pattern) {
        return pattern instanceof Date;
    }
}, {
    build: function build(pattern) {
        var patterns = pattern.map(findAndBuild);
        return function (value) {
            return value && Array.isArray(patterns) && value.length >= patterns.length && patterns.every(function (matches, i) {
                return matches(value[i]);
            });
        };
    },
    matches: function matches(pattern) {
        return Array.isArray(pattern);
    }
}, {
    build: function build(pattern) {
        return function (value) {
            return typeof value === 'string' && pattern.test(value);
        };
    },
    matches: function matches(pattern) {
        return pattern instanceof RegExp;
    }
}, {
    build: function build(pattern) {
        return function (value) {
            return value instanceof pattern;
        };
    },
    matches: function matches(pattern) {
        return pattern && pattern[Symbol.species] !== undefined;
    }
}, {
    build: function build(pattern) {
        return function (value) {
            return pattern === Boolean && typeof value === 'boolean' || pattern === Number && typeof value === 'number' || pattern === Object && (typeof value === 'undefined' ? 'undefined' : _typeof(value)) === 'object' && value || pattern === String && typeof value === 'string' || value instanceof pattern;
        };
    },
    matches: function matches(pattern) {
        return pattern === Array || pattern === Boolean || pattern === Error || pattern === Number || pattern === Object || pattern === String;
    }
}, {
    build: function build(pattern) {
        var patterns = objectEntries(pattern).map(function (_ref) {
            var _ref2 = _slicedToArray(_ref, 2);

            var key = _ref2[0];
            var pattern = _ref2[1];
            return [key, findAndBuild(pattern)];
        });
        return function (value) {
            if (!value || (typeof value === 'undefined' ? 'undefined' : _typeof(value)) !== 'object') return false;
            var entries = objectEntries(value);
            return entries.length >= patterns.length && patterns.every(function (_ref3) {
                var _ref4 = _slicedToArray(_ref3, 1);

                var patternKey = _ref4[0];
                return entries.some(function (_ref5) {
                    var _ref6 = _slicedToArray(_ref5, 1);

                    var entryKey = _ref6[0];
                    return entryKey === patternKey;
                });
            }) && patterns.every(function (_ref7) {
                var _ref8 = _slicedToArray(_ref7, 2);

                var key = _ref8[0];
                var matches = _ref8[1];
                return matches(value[key]);
            });
        };
    },
    matches: function matches(pattern) {
        return pattern && (typeof pattern === 'undefined' ? 'undefined' : _typeof(pattern)) === 'object' && pattern.__proto__ === Object.prototype;
    }
}, {
    build: function build(pattern) {
        return function (value) {
            return pattern(value);
        };
    },
    matches: function matches(pattern) {
        return typeof pattern === 'function';
    }
}];
function findPattern(pattern) {
    var foundPattern = avaliblePatterns.find(matches(pattern));
    if (!foundPattern) {
        throw Error('No matching pattern for pattern ' + pattern);
    }
    return foundPattern;
}
function pattern() {
    var patterns = [];
    var noDefaultValue = {};
    var defaultValue = noDefaultValue;
    var constructingPattern = function constructingPattern(value) {
        for (var _len = arguments.length, extra = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
            extra[_key - 1] = arguments[_key];
        }

        var matchingPattern = patterns.find(matches(value));
        if (!matchingPattern) {
            if (defaultValue === noDefaultValue) {
                throw Error('MatchError: no pattern matches value ' + value);
            }
            return typeof defaultValue === 'function' ? defaultValue.apply(undefined, [value].concat(extra)) : defaultValue;
        }
        var returnValue = matchingPattern.value;

        return typeof returnValue === 'function' ? returnValue.apply(undefined, [value].concat(extra)) : returnValue;
    };
    constructingPattern.when = function (pattern, value) {
        var foundPattern = findPattern(pattern);
        var matches = foundPattern.build(pattern);
        patterns.push({ matches: matches, value: value });
        return constructingPattern;
    };
    constructingPattern.default = function (value) {
        defaultValue = value;
        return constructingPattern.close();
    };
    constructingPattern.match = function (value) {
        for (var _len2 = arguments.length, extra = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
            extra[_key2 - 1] = arguments[_key2];
        }

        return constructingPattern.apply(undefined, [value].concat(extra));
    };
    constructingPattern.close = function () {
        return function (value) {
            for (var _len3 = arguments.length, extra = Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
                extra[_key3 - 1] = arguments[_key3];
            }

            return constructingPattern.apply(undefined, [value].concat(extra));
        };
    };
    return constructingPattern;
}
exports.pattern = pattern;
function match(value, patternBuilder) {
    var match = pattern();
    patternBuilder(match);
    return match(value);
}
exports.match = match;
exports._ = function () {
    return true;
};
exports.some = function (value) {
    return value !== null && value !== undefined;
};
exports.none = function (value) {
    return value === null || value === undefined;
};
exports.either = function () {
    for (var _len4 = arguments.length, patterns = Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
        patterns[_key4] = arguments[_key4];
    }

    patterns = patterns.map(findAndBuild);
    return function (value) {
        return patterns.some(function (matches) {
            return matches(value);
        });
    };
};
exports.range = function (start, end) {
    return function (value) {
        return start <= value && value <= end;
    };
};
exports.lt = function (number) {
    return function (value) {
        return value < number;
    };
};
exports.lte = function (number) {
    return function (value) {
        return value <= number;
    };
};
exports.gt = function (number) {
    return function (value) {
        return value > number;
    };
};
exports.gte = function (number) {
    return function (value) {
        return value >= number;
    };
};